#!/usr/bin/env python
VERSION = '0.5.1'

from corny import *
from corny.exdata import *

def main(configfile=None):
    print("Welcome to instant-PASDy %s!" % VERSION)

    # Find script path
    pasdy_path = os.path.dirname(os.path.abspath(__file__))

    # Read config
    config = read_config(configfile, chdir=True)

    # Update data (FTP, SD, ...)
    if 'update' in config:
        for source in csplit(config['update']['data_source']):
            print('| Updating data from ' + source + ' into ' + config['input']['path'])
            download(
                archive_file = config['input']['path'],        source     = source,
                server       = config['update']['FTP_server'], user       = config['update']['FTP_user'],
                pswd         = config['update']['FTP_pswd'],   ftp_prefix = config['update']['FTP_prefix'],
                ftp_suffix   = config['update']['FTP_suffix'], sd_path    = config['update']['SD_path'],
                sd_prefix    = config['update']['SD_prefix'],  sd_suffix  = config['update']['SD_suffix'])


    ####################################################
    ######  Input  #####################################
    #################################################### TODO: outsource this

    delimiter = ''
    if config['input']['path'].endswith('.zip'):
        delimiter = '/'
    pattern = config['input']['prefix']+'*'+config['input']['suffix']
    print('> Reading data from ' + config['input']['path'] + delimiter + pattern)

    archive = None
    files = []

    if config['input']['path'].endswith('.zip'):
        archive = zipfile.ZipFile(config['input']['path'], 'r')
        files = [x for x in archive.namelist() if x.endswith(config['input']['suffix']) and x.startswith(config['input']['prefix'])]

    elif os.path.isdir(config['input']['path']):
        files = glob1(config['input']['path'],pattern)
    else:
        if 'gui' in globals():
            gui.Popup("Input path is neither a folder nor a zip archive.", title='Read ERROR', non_blocking=True)
        else:
            print("! ERROR: Input path is neither a folder nor a zip archive.")
        return(None)

    if len(files) == 0:
        print(cprintf('! No matching files found.', 'red'))
        return()

    datastr = ''
    data_columns = []
    if 'input_columns' in config['input']:
        if config['input']['input_columns'] != '':
            data_columns = csplit(config['input']['input_columns'])

    i = 0
    last_progress = 0 # in %

    for filename in files:
#        print('.', end='')
        #print(filename)
        #add path again
        #filename=os.path.join(os.path.normpath(config['input']['path']),filename)
        progressbar(i-1, len(files), title='Reading', prefix=filename)
        i += 1
        progress = (i+0.01)/len(files)//0.1*10 # in %, offset 0.01 because x/x//0.1 is 9
        if progress > last_progress:
            #print("%.0f%%.." % progress, end='')
            last_progress = progress

        if datastr=='':
            if data_columns == []:
                data_columns = read_header(filename, archive)
            datastr = read(filename, archive)
        else:
            datastr += read(filename, archive) #data.append() #, verify_integrity=True)

    progressbar(i-1, len(files), title='Reading', prefix=filename, end=True)
    if not archive is None: archive.close()
    #DEBUG output raw data file
    #with open('temp.txt','a') as f:
    #    f.write(datastr)
    print('| Conversion to data frame...')
    index_col_num = 1
    if 'index_column_number' in config['input']:
        index_col_num = int(config['input']['index_column_number'])

    data = make_pandas(datastr, data_columns, index=index_col_num)

    if data is None:
        print('! Error: Cannot interprete the data. Check the column header, is this correct? '+ ','.join(data_columns))
        return()
    if len(data)==0:
        raise ValueError('No record in dataset')

    ####################################################
    ######  Basic cleaning  ############################
    ####################################################

    if cc(config, 'clean', 'bbox'):
        bbox = xsplit(config['clean']['bbox'], type=float)
    else:
        bbox = []


    lat = gc(config, 'input',      'latitude_column')
    lon = gc(config, 'input',      'longitude_column')
    alt = gc(config, 'input',      'altitude_column')
    p  =  gc(config, 'input',      'pressure_column')
    rh =  gc(config, 'input',      'humidity_column')
    T  =  gc(config, 'input',      'temperature_column')
    ah =  gc(config, 'correction', 'new_humidity_column')
    inc = gc(config, 'correction', 'new_incoming_column')
    # Thermal neutrons
    th = '_thermal'
    # Cleaned data
    clean = '_clean'
    # Corrected data
    pih = '_pih'
    # Suffix for vairable errors
    err = '_err' # followed by _sym, _low, _upp
    # Standard deviation
    std = '_std'
    # Smooth
    mov = '_mov'
    # Biomass
    bio = '_bio'
    # Urban
    urb = '_urb'
    # Road
    road = '_road'

    # check wrong column names
    for name in [p, rh, T]:
        if not name in data.columns:
            print('! Warning: %s is not a column name.' % name)
            data[name] = np.nan

    print('i   Data period from', data.index.min().strftime('%Y-%m-%d %H:%M:%S'),
                            'to', data.index.max().strftime('%Y-%m-%d %H:%M:%S'))

    # Estimate time resolution if does not exist
    tres = gc(config, 'input', 'resolution_column', alt='N1ET_sec')
    if not tres in data.columns:
        data[tres] = infer_tres(data, default=gc(config, 'input', 'resolution_default'))

    print('i Time resolution approx. %.0f sec.' % data[tres].median())

    ## Sum up neutron columns
    N =  gc(config, 'correction', 'new_neutron_column')
    data[N] = 0
    for c in xsplit(config['input']['neutron_columns']):
        data[N] += data[c]

    # Calculate error
    data[N+err] = N_err(data[N])

    # Report
    report_N(data, N, units='raw counts')

    # convert to units of cph
    data[N]     = Unit_convert(data[N], data[tres], 'cph')
    data[N+err] = Unit_convert(data[N+err], data[tres], 'cph')

    # Report
    report_N(data, N)

    # Thermal neutrons
    if config['input']['thermal_neutron_columns']:

        # sum up thermal neutrons
        data[N+th] = 0
        for c in xsplit(config['input']['thermal_neutron_columns']):
            data[N+th] += data[c]

        data[N+th+err] = N_err(data[N+th])

        # convert to units of cph
        data[N+th]     = Unit_convert(data[N+th], data[tres], 'cph')
        data[N+th+err] = Unit_convert(data[N+th+err], data[tres], 'cph')

        Nth = N+th
    else:
        Nth = None


    ####################################################
    ######  Clean/filter  ##############################
    ####################################################

    print('| Cleaning... ', end='')

    #first we clean non valid geometries


    # Clean Latlon
    if lat != '' and lat in data.columns:
        print('| Cleaning coordinates...')

        #JJ Strangely formatted coordinates occur for some Rovers (Jülich, Melbourne)

        if 'EW' in data.columns or 'NS' in data.columns:
            # Remove cardinal direction if it is attached to the coordinate (has been necessary in some configurations)
            data[lat] = pd.to_numeric(data[lat].astype(str).replace({'N':''}, regex=True), errors='coerce') # replace N if existent in column and convert to numeric
            data[lon] = pd.to_numeric(data[lon].astype(str).replace({'E':''}, regex=True), errors='coerce') # replace E if existent in column and convert to numeric
            # Convert to readable coordinate
            data[lat] = deg100min2dec( data[[lat, 'NS']] )
            data[lon] = deg100min2dec( data[[lon, 'EW']] )

        data.loc[data[lat] == 0.0, lat] = np.nan
        data.loc[data[lon] == 0.0, lon] = np.nan

        data['lon'] = data[lon]
        data['lat'] = data[lat]

        if 'interpolate_coords' in config['clean']:
            if yesno2bool(config['clean']['interpolate_coords']):
                data['lon'] = data['lon'].interpolate()
                data['lat'] = data['lat'].interpolate()

        data = data.dropna(subset=['lat'])

        if len(bbox)>0:
            data = data.query("lat > %s and lat < %s" % (bbox[2], bbox[3]))
            data = data.query("lon > %s and lon < %s" % (bbox[0], bbox[1]))
            if len(data) == 0:
                print(cprintf('! You seem to have erased the whole dataset using an unfortunate bbox.', 'red'))
                return()
        else:
        # Derive bbox from data if not set in config
            bbox = [data[lon].min(), data[lon].max(), data[lat].min(), data[lat].max()]
            print('bbox automagically craeated:', bbox)

        try:
            data['ll'] = data.apply(lambda row: tuple((row['lat'],row['lon'])), axis=1)
        except ValueError:
            raise ValueError('NO GPS Data Available, processing stopped')
        #D = D.dropna(subset=['lat'])
        # Lat lon to UTM
        utmy, utmx = latlon2utm(data.lat, data.lon)
        data['utmy'] = utmy
        data['utmx'] = utmx


        # if no geographic information available we stop processing here
        if data[lat].isna().sum()==len(data) or data[lon].isna().sum()==len(data):
            raise ValueError('NO GPS Data Available, processing stopped')

        ##replace Altitude colum with data from DEM#
        if 'altitude_raster' in config['clean'] and config['clean']['altitude_raster']:
            data = get_from_raster(data, alt, config['clean']['altitude_raster'])
            print('Altitude data replaced by values from external DEM source')



    # Get upper and lower bounds from config
    (Na,Nb)  = xsplit(config['clean']['neutron_range'], range=2, type=float)
    (pa,pb)  = xsplit(config['clean']['pressure_range'], range=2, type=float)
    (rha,rhb)= xsplit(config['clean']['humidity_range'], range=2, type=float)
    (Ta,Tb)  = xsplit(config['clean']['temperature_range'], range=2, type=float)
    if 'timeres_range' in config['clean']:
        (tra,trb)= xsplit(config['clean']['timeres_range'], range=2, type=float)

    # Define new column N+clean
    data[N+clean] = data[N]
    data[N+clean+err] = data[N+err]
    Ncleanstr = N+clean

    if Nth:
        data[N+th+clean] = data[N+th]
        data[N+th+clean+err] = data[N+th+err]

    # Drop lines option
    if config['clean']['invalid_data'] == 'drop lines':
        len_before = len(data)
        data = data.query("@Ncleanstr > %s and @Ncleanstr < %s" % (Na, Nb))
        data = data.query("@p > %s and @p < %s" % (pa, pb))
        data = data.query("@rh> %s and @rh< %s" % (rha,rhb))
        data = data.query("@T > %s and @T < %s" % (Ta, Tb))
        if 'timeres_range' in config['clean']:
            data = data.query("@tres > %s and @tres < %s" % (tra, trb))
        print("dropped %s invalid lines." % (len_before - len(data)))
    # Fill-na option
    else:
        nan_before = data.isnull().sum().sum()

        data[Ncleanstr] = find_outlier(data, Ncleanstr, lower=Na, upper=Nb, n_sigma=3, sigma=Ncleanstr+err)
        if cc(config, 'clean', 'timeres_range'):
            data[Ncleanstr] = find_outlier(data, tres, Ncleanstr, lower=tra, upper=trb)



        ################# Automatic meteo lookup (DWD)

        #start with is air humidity
        if len( data.loc[(data[rh]< rha) | (data[rh]> rhb), rh])>0:
            if config['clean']['missing_humidity'] =='DWD':
                #get subset with values out of range
                data_subset=data.loc[(data[rh]< rha) | (data[rh]> rhb)]
                #call API Functionality to DWD
                print('| Looking up air humidity from external source...', end='')
                data_subset=retrieve_dwd_data(data_subset,
                          lon=lon,lat=lat,dwd_category='air_temperature',
                          dwd_parameters=['2m_relative_humidity'],
                          no_of_nearest_stations=4)
                #finally replace the values in the dataframe
                data.loc[data_subset.index,rh] = data_subset['2m_relative_humidity']
                print(' %.1f +/- %.1f g/m3' % (data_subset['2m_relative_humidity'].mean(), data_subset['2m_relative_humidity'].std()))
            else:
                data[rh] = find_outlier(data, rh, lower=rha, upper=rhb, fill=gc(config, 'clean', 'missing_humidity'))

        #the temperature
        if len( data.loc[(data[T] < Ta ) | (data[T] > Tb) , T ] )>0:
            if config['clean']['missing_temperature'] =='DWD':
                #get subset with values out of range
                data_subset=data.loc[(data[T] < Ta ) | (data[T] > Tb)]
                #call API Functionality to DWD
                print('| Looking up air temperature from external source...', end='')
                data_subset=retrieve_dwd_data(data_subset,
                          lon=lon,lat=lat,dwd_category='air_temperature',
                          dwd_parameters=['2m_air_temperature'],
                          no_of_nearest_stations=4)
                #finally replace the values in the dataframe
                data.loc[data_subset.index,T] = data_subset['2m_air_temperature']
                print(' %.1f +/- %.1f C' % (data_subset['2m_air_temperature'].mean(), data_subset['2m_air_temperature'].std()))
            else:
                data[T] = find_outlier(data, T, lower=Ta, upper=Tb, fill=gc(config, 'clean', 'missing_temperature'))

        #finally the pressure

        if len(data.loc[(data[p] < pa ) | (data[p] > pb) , p ])>0:
            if config['clean']['missing_pressure'] =='DWD':
                print('correct missing values using dwd database')
                #get subset with values out of range
                data_subset=data.loc[(data[p] < pa ) | (data[p] > pb)]
                #call API Functionality to DWD
                print('| Looking up air pressure from external source...', end='')
                data_subset=retrieve_dwd_data(data_subset,
                          lon=lon,lat=lat,dwd_category='pressure',
                          dwd_parameters=['air_pressure_nn'],
                          no_of_nearest_stations=10)
                # we use the barometric formular to correct pressure data
                #p=p0*(1-0.0065h/(T+0.0065h+273.15))^5.257
                data_subset['air_pressure_local']=data_subset['air_pressure_nn']*(1-(0.0065*data_subset[alt])/(data_subset[T]+0.0065*data_subset[alt]+273.15))**5.257
                #finally replace the values in the dataframe
                data.loc[data_subset.index,p] = data_subset['air_pressure_local']
                #remove 0 values, as they come from defect pressure measurements
                data=data[data[p]!=0]
                data_subset=data_subset[data_subset[p]!=0]
                print(str(sum(data[p]==0)), 'of ', str(len(data)),'records removed due invalid dwd pressure data')
                print(' %.1f +/- %.1f hPa' % (data_subset['air_pressure_local'].mean(), data_subset['air_pressure_local'].std()))
                if len(data)==0:
                    raise ValueError('No entries with valid pressure from dwd left')
            else:
                data[p] = find_outlier(data, p, lower=pa, upper=pb, fill=gc(config, 'clean', 'missing_pressure'))


        #result
        print("replaced %s values by NaN." % (data.isnull().sum().sum() - nan_before))


    # Interpolate over nans
    if yesno2bool(config['clean']['interpolate_neutrons']):
        data[N+clean]     = data[N+clean].interpolate()
        data[N+clean+err] = data[N+clean+err].interpolate()
        if Nth:
            data[N+th+clean]     = data[N+th+clean].interpolate()
            data[N+th+clean+err] = data[N+th+clean+err].interpolate()

    report_N(data, N+clean)

    ####################################################
    ######  Soil #######################################
    ####################################################

    print('> Sampling land data... ', end='')
    #depending on Option for soil data retrieval we add soil data
    #print('Retrieve soil properties from constant values')
    #add static data -- this is the default and also the basis
    data['bd']  = float(config['conversion']['bulk_density'])
    data['org'] = float(config['conversion']['soil_org_carbon'])
    data['lw']  = float(config['conversion']['lattice_water'])

    # add from raster datasets
    if 'soil_data_source' in config['conversion'] and config['conversion']['soil_data_source']=='raster':
        print('Retrieve soil properties from static maps')
        if lat in data.columns:
            if config['conversion']['bulk_density_raster']:
                data = get_from_raster(data, 'bd',   config['conversion']['bulk_density_raster'])
            if config['conversion']['clay_content_raster']:
                data = get_from_raster(data, 'clay', config['conversion']['clay_content_raster'])

            if config['conversion']['soil_org_carbon_raster']:
                data = get_from_raster(data, 'org',  config['conversion']['soil_org_carbon_raster'])

        #convert lattice water from clay
        data['lw']  = lw_from_clay(data['clay'], method='Greacen et al. 1981')

    # add from soilgrids
    if 'soil_data_source' in config['conversion'] and config['conversion']['soil_data_source']=='soilgrids':
        print('Retrieve soil properties from SoilGrids API')
        soil_data=retrieve_soilgrid_data(data.copy(),lon=lon,lat=lat,
                                        soilgridlrs=['bdod','clay','soc'],
                                        soil_layerdepths=['0-5cm','5-15cm','15-30cm','30-60cm'],
                                        raster_res=(250, 250),
                                        statistical_metric=['mean'],
                                        all_touched=True
                                        )
        #add the soil data to the dataset
        data['org']=soil_data['org'] /10 # new units
        data['bd']=soil_data['bd']
        data['clay']=soil_data['clay']

        #convert lattice water from clay
        data['lw']  = lw_from_clay(data['clay'], method='Greacen et al. 1981')

    #convert organic water equivalent from organic matter content
    data['owe'] = owe_from_corg(data['org'], method='Franz et al. 2015')

    #add landuse data
    if lat in data.columns:
        if config['conversion']['land_use_raster']:
            data = get_from_raster(data, 'luse', config['conversion']['land_use_raster'])

    print('bulk dens. (%.2f g/cm3), lattice w. (%.2f g/g), org.we. (%.2f g/g)' % (data.bd.mean(), data.lw.mean(), data.owe.mean()))


    ####################################################
    ######  Incoming  ##################################
    ####################################################

    print(cprintf('> Reading incoming radiation data from ' + config['correction']['NM_path'], 'yellow'), end='')
    M = NM(config=config).get(start=data.index.min().replace(tzinfo=pytz.UTC), end=data.index.max().replace(tzinfo=pytz.UTC))

    data.index = data.index.tz_localize(None)
    data.index = data.index.tz_localize('UTC')
    if isinstance(M.data, pandas.DataFrame):
        M.data.index = M.data.index.tz_localize(None)
        M.data.index = M.data.index.tz_localize('UTC')
        M.data = M.data.reindex(data.index, method='nearest', limit=1, tolerance=timedelta(seconds=M.resolution_min*60/2-1)).interpolate()
        data[inc] = M.data
        data.loc[data[inc] < data[inc].mean()/1.2, inc] = np.nan
        data.loc[data[inc] > data[inc].mean()*1.2, inc] = np.nan
    else:
        data[inc] = float(config['correction']['incoming_ref'])

    ####################################################
    ######  pih  #######################################
    ####################################################

    print("\n| Corrections... ", end='')

    data[ah] = abs_humidity(data[rh], data[T]) #6.112*np.exp(17.67*data[T]/(243.5+data[T]))/(273.15 + data[T]) * 2.1674 * data[rh]

    data['correct_h'] = C_humid( data[ah], gc(config, 'correction', 'humidity_ref', dtype=float),
                                           gc(config, 'correction', 'alpha', dtype=float),
                                           method=gc(config, 'correction', 'humidity_method'))

    data['correct_p'] = C_press( data[p],  gc(config, 'correction', 'pressure_ref', dtype=float),
                                           gc(config, 'correction', 'beta', dtype=float),
                                           method=gc(config, 'correction', 'pressure_method'),
                                           inclination = gc(config, 'correction', 'inclination', dtype=float))

    #np.exp((data[p] - float(config['correction']['pressure_ref'])) / float(config['correction']['beta']))
    data['correct_i'] = C_inc( data[inc],  gc(config, 'correction', 'incoming_ref', dtype=float),
                                           gc(config, 'correction', 'gamma', dtype=float),
                                           method=gc(config, 'correction', 'incoming_method'),
                                           Rc = gc(config, 'correction', 'Rc', dtype=float),
                                           Rc_NM = gc(config, 'correction', 'Rc_NM', dtype=float))


    data[N+clean+pih]     = data[N+clean]     * data['correct_p'] * data['correct_i'] * data['correct_h']
    data[N+clean+pih+err] = data[N+clean+err] * data['correct_p'] * data['correct_i'] * data['correct_h']

    if Nth:
        data[N+th+clean+pih]     = data[N+th+clean]     * data['correct_p'] * data['correct_i'] * data['correct_h']
        data[N+th+clean+pih+err] = data[N+th+clean+err] * data['correct_p'] * data['correct_i'] * data['correct_h']
        Nthfinal = N+th+clean+pih

    print("for pressure (%+.0f%%), abs. humidity (%+.0f%%), incoming radiation (%+.0f%%)" % \
        (100*(data['correct_p'].mean()-1), 100*(data['correct_h'].mean()-1), 100*(data['correct_i'].mean()-1)))

    Nfinal = N+clean+pih
    report_N(data, N+clean+pih, correction=['correct_p','correct_i','correct_h'])
    #If no data remain with valid corrected data, we stop processing
    if data[N+clean+pih].isna().sum()==len(data):
        raise ValueError('No record with valid meteo OR LAT/LON OR0 Neutron Range exists, stop further processing')


    ####################################################
    ######  POI  #######################################
    ####################################################

    if 'poi_json' in config['conversion'] and config['conversion']['poi_json']: # or config['conversion']['poi_lat']:
        print('| Evaluating points of interest...')
        poi = '_poi'
        POI = read_geojson(config['conversion']['poi_json']) # 'data/Selke.geojson'
        utmy, utmx = latlon2utm(POI.lat, POI.lon)
        POI['utmy'] = utmy
        POI['utmx'] = utmx
        iax = -1
        max_radius = config['conversion']['poi_radius']

        data[Nfinal+'_poi'] = data[Nfinal]
        data[Nfinal+'_poi_err'] = data[Nfinal+err]
        Nfinal += '_poi'

        for i, row in POI[POI.type=='UFZ'].iterrows():
            data['distance_poi'] = np.sqrt((data.utmx-row['utmx'])**2+(data.utmy-row['utmy'])**2)
            iax += 1
            data.loc[data.distance_poi<max_radius, Nfinal] = data.loc[data.distance_poi<max_radius, Nfinal].mean()
            data.loc[data.distance_poi<max_radius, Nfinal+err] = data.loc[data.distance_poi<max_radius, Nfinal+err].mean() / np.sqrt(data.loc[data.distance_poi<max_radius, Nfinal+err].count())
            print('i   %20s %.0f' % (row['label'], data.loc[data.distance_poi<max_radius, Nfinal].mean()))
    else:
        poi = ''

    ####################################################
    ######  Urban  #####################################
    ####################################################

    data['correct_urban'] = 1

    if 'luse_str' in data.columns and 'land_use_urban_corr' in config['conversion'] and config['conversion']['land_use_urban_corr'] != '':
        if config['conversion']['land_use_urban_corr'] == 'nan':
            data.loc[data.luse_str=='urban', 'correct_urban'] = np.nan
        else:
            data.loc[data.luse_str=='urban', 'correct_urban'] = float(config['conversion']['land_use_urban_corr'])

        data[Nfinal+urb] = data[Nfinal] * data['correct_urban']
        data[Nfinal+urb+err] = data[Nfinal+err] * data['correct_urban']
        Nfinal += urb
        report_N(data, Nfinal, 'correct_urban')
    else:
        urb = ''

    ####################################################
    ######  Bio  #######################################
    ####################################################
    ## vegetation correct
    ## assume neutron counts were reduced by 10% for every measurement in a forest
    data['correct_bio'] = 1
    data['correct_bio_err'] = 0
    if 'luse_str' in data.columns and 'land_use_forest_corr' in config['conversion'] and config['conversion']['land_use_forest_corr'] != '':
        if config['conversion']['land_use_forest_corr'] == 'nan':
            data.loc[data.luse_str=='forest', 'correct_bio'] = np.nan
        else:
            data.loc[data.luse_str=='forest', 'correct_bio'] = float(config['conversion']['land_use_forest_corr'])

        print("| Correction for biomass (%+.0f%%)." % (100*(data.loc[data.luse_str=='forest', 'correct_bio'].mean()-1)))
        # well do that after road correction
        #data[Nfinal+'_bio'] = data[Nfinal] * data['correct_bio']
        #data[Nfinal+'_bio'+err] = data[Nfinal+err] * data['correct_bio']
        #Nfinal += '_bio'
        #report_N(data, Nfinal, 'correct_bio')

    else:
        bio = ''
        #print('.')


    ####################################################
    ######  Road  ######################################
    ####################################################

    data['correct_road'] = 1
    if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None' and lat in data.columns:

        print('| Correction for road effect...')
        # Load road properties (goes to config?)

        #print('|   Loading road type correction properties...')
        print(cprintf('>   %s' % config['conversion']['road_type_table'], 'yellow'))
        Roads = pd.read_csv(config['conversion']['road_type_table'], sep=r'\s+', skipinitialspace=True, comment='#')

        ####Now we decide whether we get the road_network_Data####
        ####from osmnx or fromroverweb Overpass API####
        if 'road_network_source' in config['conversion'] and config['conversion']['road_network_source']=='osm-nx':
            print('retrieve road network information from osmnx')
            # Load road network
            #print('>   ', end='') # Graph prints a message.
            road_network_file = config['conversion']['road_network_file']

            # Load nearest road types
            road_type_file = os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] +'.roads' #configfile[0:configfile.index('.cfg')]

            print('|   Loading road network... ', end='')
            if os.path.exists(road_type_file):
                # Read from file
                #print('|   Loading nearest road types...')
                print(cprintf('\n>   %s' % road_type_file, 'yellow'))
                roadsdf = pd.read_csv(road_type_file, index_col=0, parse_dates=True, infer_datetime_format=True)
                data['road'] = roadsdf
            else:
                #if config['conversion']['road_network_source'] == 'osmnx':
                R = Graph('roads', bbox, file=road_network_file, report_save='<   Saved')
                R.get(save=True) # read from file, or download and save

                if len(R.data)<=0:
                    print('i   Road network is empty!')
                else:
                    # Calculate
                    print('|   Searching nearest roads...')
                    # invoke progress bar feature
                    from tqdm import tqdm
                    tqdm.pandas()

                    data['road'] = data.progress_apply(lambda row: R.on_road_type(point=row['ll']), axis=1)

                    #print('    Set priorities...')
                    for i, row in data.iterrows():
                        row = row.copy()
                        if isinstance(row.road, list):
                            if 'primary' in row.road:
                                data.loc[i, 'road'] = 'primary'
                            elif 'track' in row.road:
                                data.loc[i, 'road'] = 'track'
                        #print(D.loc[i, 'road'], end=' ')
                    #print('')

                    # Save backup
                    data['road'].to_csv(road_type_file)
                    file_save_msg(road_type_file, 'Nearest roads', end="\n")

        elif 'road_network_source' in config['conversion'] and config['conversion']['road_network_source']=='osm-overpass':
            print('| Overpass')
            print('retrieve road network information from overpass API')
            data=retrieve_osm_overpass_data(data, lon=lon,lat=lat,crcl_radius=50,
                               no_of_clusters=None,queryfeatures={'way': ['highway']},
                               output_col_name='road')
        # Set road values
        data['road_moisture'] = np.nan
        data['road_width'] = 0.0
        if 'road' in data.columns:
            for i, row in Roads.iterrows():
                row = row.copy()
                data.loc[data.road==row['category'], 'road_moisture'] = row['moisture']
                data.loc[data.road==row['category'], 'road_width']    = row['width']
        else:
            print('|   Applying constant road parameters theta=%s, width=%s' % (config['conversion']['road_moisture'], config['conversion']['road_width']))
            data['road_moisture'] = float(config['conversion']['road_moisture'])
            data['road_width']    = float(config['conversion']['road_width'])
        #pdb.set_trace()
        # correct for biomass to avoid smearing out
        data[Nfinal+'_temp'] = data[Nfinal] * data['correct_bio']
        data = moving_avg(data, Nfinal+'_temp', window=config['clean']['smooth'], err=False, std=False)
        # back to Nfinal, but mavged
        data[Nfinal+'_temp_mov'] /= data['correct_bio']
        # N2SM first order
        if config['conversion']['N0']:
            N0 = config['conversion']['N0']
        else:
            N0 = data.Nfinal.max()*1.1
            print('! N0 is not defined, but we need it for road correction. Guessing %.0f cph.' % N0)
        data['sm_temp'] = N2SM_Desilets(data[Nfinal+'_temp_mov'], N0, bd=data.bd, lw=data.lw, owe=data.owe,
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        # C_road
        data['correct_road'] = 1/(1 + 0.42*(1-np.exp(-0.5*data.road_width)) * (1.06-4*data.road_moisture-(0.16 + data.road_moisture)/(0.39 + data['sm_temp'])))
        data.loc[(data['correct_road'] > 1), 'correct_road'] = 1
        #data.loc[data.road_width==0, 'correct_road'] = 1
        data.correct_road = data.correct_road.fillna(1)
        print("i   ...road effect (%+.0f%%)." % (100*(data['correct_road'].mean()-1)))

        # tidy up temp columns
        #data.drop([Nfinal+'_temp', 'sm_temp'], axis=1, inplace=True)

    else:
        road = ''


    # Set N_bio
    data[Nfinal+bio]          = data[Nfinal]         * data['correct_bio']

    if 'land_use_forest_corr_err' in config['conversion']:
        data['correct_bio_err'] = float(config['conversion']['land_use_forest_corr_err'])
    data[Nfinal+bio+err] = errpropag_factor(data, Nfinal, 'correct_bio', err)
    #data[Nfinal+bio+err]      = data[Nfinal+err]     * data['correct_bio']

    Nfinal += bio
    report_N(data, Nfinal, 'correct_bio')

    if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None'  and lat in data.columns:

        # set N_road
        data[Nfinal+road]     = data[Nfinal]     * data['correct_road']

        data['correct_road_err'] = float(config['conversion']['road_correction_err'])
        data[Nfinal+road+err] = errpropag_factor(data, Nfinal, 'correct_road', err)
        #data[Nfinal+road+err] = data[Nfinal+err] / data['correct_road']

        Nfinal += road
        report_N(data, Nfinal, 'correct_road')

    #if config['conversion']['road_moisture'] and float(config['conversion']['road_width']) > 0 and not 'correct_road' in data.columns:
     #           print("| Corrections for road effect... ", end='')
                #if   config['correction']['road_material'] == 'stone':    road_theta = 0.04
                #elif config['correction']['road_material'] == 'concrete': road_theta = 0.07
                #elif config['correction']['road_material'] == 'asphalt':  road_theta = 0.10
      #          road_theta = float(config['conversion']['road_moisture'])


    neutron_columns = [Nfinal]
    if Nth: neutron_columns += [N+th, N+th+clean, N+th+clean+pih]

    ####################################################
    ######  Aggregate ##################################
    ####################################################

    if cc(config,'clean','aggregate'):

        # keep a copy
        data_orig = data.copy()

        # Aggregate
        data = aggregate(data,  gc(config,'clean','aggregate'),
                         func = gc(config,'clean','aggregate_func', alt='mean'),
                         set_std = neutron_columns, adapt_err = neutron_columns,
                         verbose = True)
        # Report
        report_N(data, Nfinal)

    else:
        data_orig = data

    ####################################################
    ######  Smooth  ####################################
    ####################################################

    if config['clean']['smooth']:
        print('| Smooth data by a moving average over '+config['clean']['smooth']+' time steps...')

        data = moving_avg(data, Nfinal, window=config['clean']['smooth'])
        Nfinal += mov
        report_N(data, Nfinal)

        #window = int()
        #rollobj = data[N+clean+pih].rolling(window, center=True)
        #data[N+clean+pih+mov]     = rollobj.mean()
        #data[N+clean+pih+mov+std] = rollobj.std()
        #data[N+clean+pih+mov+err] = data[N+clean+pih+err] * 1/np.sqrt(window)

        if Nth:
            data = moving_avg(data, Nthfinal, window=config['clean']['smooth'])
            Nthfinal += mov

            #rollobj = data[N+th+clean+pih].rolling(window, center=True)
            #data[N+th+clean+pih+mov]     = rollobj.mean()
            #data[N+th+clean+pih+mov+std] = rollobj.std()
            #data[N+th+clean+pih+mov+err] = data[N+th+clean+pih+err] * 1/np.sqrt(window)
    else:
        mov = ''


    ####################################################
    ######  Footprint length ###########################
    ####################################################

    if 'Speed_kmh' in data.columns:
        tres = data.index.to_series().dropna().diff().median().total_seconds()
        data['footprint_length'] = data['Speed_kmh']/3.6 * tres
        print('i   Average footprint length is %.0f m.' % data.loc[data['footprint_length'] > 2.0, 'footprint_length'].mean())


    ##JJ smooth with nearest neighbors
    if config['clean']['number_locations']:
        print('| Smooth data with '+config['clean']['number_locations']+' nearest neighbors.')
        data = data.dropna(subset=['LatDec', 'LongDec']) # drop rows without Latitude or Longitude information
        maxdist, RecNums = Neighbours(data[lat], data[lon], int(config['clean']['number_locations']), data)

        data['maxdistance'] = maxdist
        data['neighbors_RecordNum'] = RecNums.tolist()

        test = []
        Recind = []
        res = []
        result = []
        resTh = []
        result_Th = []

        for i in range(len(data)): # store the nearest recordnumber in neighbors
            test = data['RecordNum'].iloc[data['neighbors_RecordNum'][i]].values
            Recind.append(test.tolist())

        data['neighbors_RecordNum'] = Recind

        for i in range(len(data)): # calculate average N from neighbors index
            res = data[Nfinal][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
            result.append(res)
            if Nth:
                resTh = data[Nthfinal][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
                result_Th.append(resTh)

        Nfinal += '_neighbors'
        Nthfinal += '_neighbors'
        data[Nfinal] = result
        data[Nthfinal] = result_Th


    ####################################################
    ######  Select Period  #############################
    ####################################################

    if config['clean']['start'] or config['clean']['end']:
        if config['clean']['start']:
            start = dateutil.parser.parse(config['clean']['start']).replace(tzinfo=pytz.UTC)
        else:
            start = data.index.min()
        if config['clean']['end']:
            end = dateutil.parser.parse(config['clean']['end']).replace(tzinfo=pytz.UTC)
        else:
            end = data.index.max()

        # cut
        data = data.loc[start:end]
        # cut also the raw data
        data_orig = data_orig.loc[start:end]

    # guess a original data resolution to plot. #TODO Make this configurable later.
    tres_temp = data_orig.index.to_series().dropna().diff().median().total_seconds()
    if tres_temp < 60*10: # 10 minutes
        data_orig = data_orig.resample('10S').mean()
    else:
        data_orig = data_orig.resample('1H').mean()

    start_str = data.index.min().strftime('%Y%m%d%H')
    end_str = data.index.max().strftime('%Y%m%d%H')
    end_str = '-'+end_str if start_str != end_str else ''
    output_basename = os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] + '-' \
                      + start_str + end_str #config['input']['prefix']

    ####################################################
    ######  Soil Moisture  #############################
    ####################################################

    if config['conversion']['N2sm_method'] != 'None':

        # Calibrate
        N0 = config['conversion']['N0']
        calibration = False
        if not N0:
            print('| Calibration... ', end='')
            calibration = True
            start = dateutil.parser.parse(config['conversion']['campaign_start']).replace(tzinfo=pytz.UTC)
            end = dateutil.parser.parse(config['conversion']['campaign_end']).replace(tzinfo=pytz.UTC)
            datac = data.copy()
            datac = datac.loc[start:end]
            Nc = datac[Nfinal].mean()
            Nc_err = datac[Nfinal].std()
            print('(sm=%.3f, N=%i, bd=%.3f, lw=%.3f, owe=%.3f)' % (float(config['conversion']['measured_sm']), Nc, data.bd.mean(), data.lw.mean(), data.owe.mean()))
            N0     = Calibrate_N0_Desilets(Nc, float(config['conversion']['measured_sm']), bd=data.bd.mean(), lw=data.lw.mean(), owe=data.owe.mean(),
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))       #                Nc/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))
            N0_err = Calibrate_N0_Desilets(Nc+Nc_err, float(config['conversion']['measured_sm']), bd=data.bd.mean(), lw=data.lw.mean(), owe=data.owe.mean(),
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2'])) - N0  #/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))

            print(cprintf('i   N0 = %.0f +/- %.0f cph' % (N0, abs(N0_err)), 'cyan')) # todo: make a report using N0 as column
        else:
            N0 = float(N0)

        # Conversion
        sm = config['conversion']['new_moisture_column']
        sm_neighbors = sm+'_neighbors'

        #while True:
        smv = sm+'_vol'
        smg = sm+'_grv'

        print('| Conversion to water content (N0=%i, bd=%.3f, lw=%.3f, owe=%.3f)... ' % (N0, data.bd.mean(), data.lw.mean(), data.owe.mean()))
        data[smg] = N2SM_Desilets(data[Nfinal], N0, bd=1, lw=data.lw, owe=data.owe,
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        data[smv] = N2SM_Desilets(data[Nfinal], N0, bd=data.bd, lw=data.lw, owe=data.owe,
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))

        # same for orig
        data_orig[smv] = N2SM_Desilets(data_orig[N+clean+pih], N0, bd=data_orig.bd, lw=data_orig.lw, owe=data_orig.owe,
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))

        data[smg+err]        = dtheta_stdx(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        data[smg+err+'_low'] = dtheta_low(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        data[smg+err+'_upp'] = dtheta_upp(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        #report_SM(data, smg)

        data['bd_err'] = 0
        if 'bulk_density_err' in config['conversion']:
            data['bd_err'] = float(config['conversion']['bulk_density_err'])
        data[smv+err]        = errpropag_factor(data, 'bd', smg, err=err)                  #data.bd * data[smg+err]        + data[smg]        *
        data[smv+err+'_low'] = errpropag_factor(data, 'bd', smg, err=err, err2=err+'_low') #data.bd * data[smg+err+'_low'] + data[smg+err+'_low'] * float(config['conversion']['bulk_density_err'])
        data[smv+err+'_upp'] = errpropag_factor(data, 'bd', smg, err=err, err2=err+'_upp') #data.bd * data[smg+err+'_upp'] + data[smg+err+'_upp'] * float(config['conversion']['bulk_density_err'])
        if data[smv+err+'_low'].mean() > 0:
            data[smv+err+'_low'] *= -1
        #report_SM(data, smv)

    ####################################################
    ######  Clean soil moisture  #######################
    ####################################################

        validlinesg = len(data[smg].dropna())
        validlinesv = len(data[smv].dropna())
        (a,b) = xsplit(config['clean']['sm_range'], type=float)

        if config['clean']['invalid_data'] == 'drop':
            data = data.query("@"+smg+" > %s and @"+smg+" < %s" % (a,b))
            data = data.query("@"+smv+" > %s and @"+smv+" < %s" % (a,b))
        else:
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err+'_low'] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err+'_upp'] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg] = np.nan
            data.loc[(data[smg]-data[smg+err] < a), smg+err] = data[smg]
            data.loc[(data[smg]+data[smg+err] > 1), smg+err] = 1-data[smg]
            data.loc[(data[smg]+data[smg+err+'_low'] < a), smg+err+'_low'] = -data[smg]
            data.loc[(data[smg]+data[smg+err+'_upp'] > b), smg+err+'_upp'] = 1-data[smg]

            data.loc[(data[smv] < a) | (data[smv] > b), smv+err] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv+err+'_low'] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv+err+'_upp'] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv] = np.nan
            data.loc[(data[smv]-data[smv+err] < a), smv+err] = data[smv]
            data.loc[(data[smv]+data[smv+err] > b), smv+err] = 1-data[smv]
            data.loc[(data[smv]+data[smv+err+'_low'] < a), smv+err+'_low'] = -data[smv]
            data.loc[(data[smv]+data[smv+err+'_upp'] > b), smv+err+'_upp'] = 1-data[smv]
            #if sm_neighbors in data.columns:
            #    data.loc[(data[sm_neighbors] < a) | (data[sm_neighbors] > b), sm_neighbors] = np.nan

        data[sm] = data[smg]
        data[sm+'_%'] = data[sm] * 100

    #print('dropped %i smg values out of %i' % (validlinesg-len(data[smg].dropna()), validlinesg))
    print('i Dropped %i out of %i values (out of range) from %s.' % (validlinesv-len(data[smv].dropna()), validlinesv, smv))

    report_SM(data, smg, units='grv.%%')
    report_SM(data, smv, units='vol.%%')
    #print('%.3f +/- %.3f -%.3f +%.3f' % (data[smv].mean(), data[smv+err].mean(), data[smv+err+'_low'].mean(), data[smv+err+'_upp'].mean()))

    ####################################################
    ######  Depth  #####################################
    ####################################################

    data['footprint_depth'] = D86(data[smv], data.bd)
    data['footprint_depth'] = data['footprint_depth'].replace(np.inf, np.nan)
    print('i   Footprint depth  D = %3.0f +/- %2.0f cm.' % (data.footprint_depth.dropna().mean(), data.footprint_depth.dropna().std()))

    ####################################################
    ######  Footprint  #################################
    ####################################################

    footprint_data = np.loadtxt(os.path.join(pasdy_path ,'corny/footprint_radius.csv'))
    def get_footprint(sm, h, p):
        if np.isnan(sm) or np.isnan(h) or np.isnan(p): return(np.nan)
        if sm <  0.01: sm =  0.01
        if sm >  0.49: sm =  0.49
        if h  > 30   : h  = 30
        #print(sm, h, p, int(round(100*sm)), int(round(h)))
        return(footprint_data[int(round(100*sm))][int(round(h))] * 0.4922/(0.86-np.exp(-p/1013.25)))

    data['footprint_radius'] = data.apply(lambda row: get_footprint( row[smv], row[ah], row[p] ), axis=1)
    print('i   Footprint radius R = %3.0f +/- %2.0f m.' % (data.footprint_radius.dropna().mean(), data.footprint_radius.dropna().std()))

    data['footprint_volume'] = (data['footprint_depth']+D86(data[smv], data.bd,data['footprint_radius']))*0.01*0.47*data['footprint_radius']**2*3.141 /1000
    print('i   Footprint volume V = %3.0f +/- %2.0f x10^3 m^3.' % (data.footprint_volume.dropna().mean(), data.footprint_volume.dropna().std()))

    # 0.44 (dry) ..0.5 (wet) is roughly the average D over radii

    ####################################################
    ######  Time Resolution  ###########################
    ####################################################

    tres_orig = data_orig.index.to_series().dropna().diff().median().total_seconds()
    tres_orig_str = ''
    if   tres_orig/86400 >= 1: tres_orig_str = '%.0fd' % (tres_orig/86400)
    elif tres_orig/3600  >= 1: tres_orig_str = '%.0fh' % (tres_orig/3600)
    elif tres_orig/60    >= 1: tres_orig_str = '%.0fmin' % (tres_orig/60)
    else                     : tres_orig_str = '%.0fsec' % (tres_orig)

    tres = data.index.to_series().dropna().diff().median().total_seconds()
    tres_str = ''
    if   tres/86400 >= 1: tres_str = '%.0fd' % (tres/86400)
    elif tres/3600  >= 1: tres_str = '%.0fh' % (tres/3600)
    elif tres/60    >= 1: tres_str = '%.0fmin' % (tres/60)
    else                : tres_str = '%.0fsec' % (tres)

    ####################################################
    ######  POI  #######################################
    ####################################################

    if cc(config, 'conversion', 'poi'): #'poi' in config['conversion'] and config['conversion']['poi']: # or config['conversion']['poi_lat']:
        print('| Evaluating points of interest...')
        #poi = '_poi'
        #POI = read_geojson(config['conversion']['poi_json']) # 'data/Selke.geojson'
        POI = pandas.read_csv(config['conversion']['poi'], skipinitialspace=True)
        utmy, utmx = latlon2utm(POI.lat, POI.lon)
        POI['utmy'] = utmy
        POI['utmx'] = utmx
        iax = -1
        max_radius = int(config['conversion']['poi_radius'])

        poi_p = []
        poi_rh = []
        poi_T = []
        poi_h = []
        poi_N = []
        poi_N_std = []
        poi_N_err = []
        poi_bio = []
        poi_bd = []
        poi_lw = []
        poi_org = []
        poi_smg = []
        poi_smg_err_low = []
        poi_smg_err_upp = []
        poi_smv = []
        poi_smv_err_low = []
        poi_smv_err_upp = []
        poi_depth = []

        for i, row in POI.iterrows():
            iax += 1
            data['distance_poi'] = np.sqrt((data.utmx-row['utmx'])**2+(data.utmy-row['utmy'])**2)
            data_poi = data[data.distance_poi < max_radius].copy()
            err_norm = np.sqrt(len(data_poi))

            #data.loc[data.distance_poi<max_radius, Nfinal] = data.loc[data.distance_poi < max_radius, Nfinal].mean()
            #data.loc[data.distance_poi<max_radius, Nfinal+err] = data.loc[data.distance_poi<max_radius, Nfinal+err].mean() / np.sqrt(data.loc[data.distance_poi<max_radius, Nfinal+err].count())

            # Printout
            # TODO: devision by err_norm for sm is an approximation, the error should be propagated properly
            print('POI %10s (%4i points) N=%5.0f +/-%4.0f cph, sm=%6.3f %5.3f+%5.3f vol.%%, depth=%3.0f +/-%2.0f cm' % (row['ID'],
                len(data_poi), data_poi[Nfinal].mean(), data_poi[Nfinal+err].mean() / err_norm,
                data_poi[smv].mean(), data_poi[smv+err+'_low'].mean() / err_norm, data_poi[smv+err+'_upp'].mean() / err_norm,
                data_poi['footprint_depth'].mean(), data_poi['footprint_depth'].std()))

            # All to single CSVs?
            if yesno2bool(config['conversion']['poi_output_fulldata']):
                datetime_format = re.sub(r'(\w)', r'%\1', config['output']['CSV_datetime_format'])
                data.to_csv(output_basename+'_POI-'+row['ID']+'.csv',
                    date_format = datetime_format,
                    sep = config['output']['CSV_column_sep'].strip('\"'),
                    decimal = config['output']['CSV_decimal'],
                    float_format = '%' + config['output']['CSV_float_format'],
                    na_rep = config['output']['CSV_NaN'])
        poi = ''


    ####################################################
    ######  Output  ####################################
    ####################################################

    if yesno2bool(config['output']['make_CSV']):
        print('| CSV data preparation... ')
        datetime_format = re.sub(r'(\w)', r'%\1', config['output']['CSV_datetime_format'])
        try:
            data.to_csv(output_basename+'.csv',
                date_format = datetime_format,
                sep = config['output']['CSV_column_sep'].strip('\"'),
                decimal = config['output']['CSV_decimal'],
                float_format = '%' + config['output']['CSV_float_format'],
                na_rep = config['output']['CSV_NaN'])
        except:
            print('! ERROR: Cannot write, probably permission denied.')

        file_save_msg(output_basename + '.csv', "Data", end="\n")


    # Figures
    if yesno2bool(config['output']['make_PDF']):

        print('| Plotting... ', end='')
        figure_name = output_basename + '.pdf'

        #print(data[N].to_string())
        ## With PdfPages: Create PDF on enter and close PDF on exit
        ## see https://matplotlib.org/gallery/misc/multipage_pdf.html
        with matplotlib.backends.backend_pdf.PdfPages(figure_name) as fig:

            Fig.time_format = config['output']['PDF_time_format']

            # Cover
            if 'info' in config:

                stamp = 'CoRNy v%s\n%s' % (VERSION, datetime.now().strftime('%b %d, %Y\n%H:%M UTC'))
                with Fig(fig, title='Data Report', size=(10.4,5.5)) as ax:
                    plt.axis('off')
                    plt.annotate(stamp, (0.99,0.99), fontsize=7, ha='right')
                    plt.annotate(config['info']['title'].encode('latin-1').decode('utf-8'), (0.05,0.9), fontsize=16)
                    plt.annotate(config['info']['subtitle'].encode('latin-1').decode('utf-8'), (0.05,0.82), fontsize=9)
                    plt.annotate('Contact: '+config['info']['contact'].encode('latin-1').decode('utf-8'), (0.05,0.78), fontsize=9)
                    #import matplotlib.image as mpimg
                    if os.path.exists(config['info']['cover']):
                        logo = plt.imread(config['info']['cover'])
                        ax.figure.figimage(logo, 0.3, 0.5, zorder=1)

            # Neutrons, label='%s resolution, $\pm$%0.f smoothing window' % (tres_str, int(config['clean']['smooth'])/2))
            if config['clean']['smooth']:
                title_str = 'Neutron time series (%s aggregation, $\pm$%0.f smoothing window)' % (tres_str, int(config['clean']['smooth'])/2)
            else:
                title_str = 'Neutron time series (%s aggregation)' % (tres_str)

            with Fig(fig, title=title_str, ylabel='Neutrons (in cph)') as ax:
                if not data[N].isnull().all():
                    plt.plot(data[N],           label="N raw counts in cph",     ls='', marker='o', ms=3, markeredgewidth=0, color='red')
                    plt.plot(data[N+clean],     label="N cleaned", ls='', marker='o', ms=3.5, markeredgewidth=0, color='whitesmoke')
                    #plt.plot(data[N+clean+pih], label="N_cc_pih", ls='', marker='o', ms=3,  color='C0', alpha=0.6, markeredgewidth=0) # , drawstyle='steps'
                    #plt.plot(data[N+clean+pih+urb+bio], label="N_cc_pih_ub", ls='', marker='o', ms=3,  color='C1', alpha=0.6, markeredgewidth=0)
                    if Nfinal == N+clean+pih:
                        plt.plot(data[N+clean+pih],                                lw=2, label="N c&corr press,incom,humid", color='black')
                    else:
                        plt.plot(data[N+clean+pih], ls='', marker='o', ms=3.5, markeredgewidth=0, label="N c&corr press,incom,humid", color='lightgrey')
                    #if urb != '': plt.plot(data[N+clean+pih+urb],              lw=4, label="N c&corr press,incom,humid,urban", color='C6')
                    #if bio != '': plt.plot(data[N+clean+pih+urb+bio],          lw=3, label="N c&corr press,incom,humid,urban,bio", color='C2')
                    #if road!= '': plt.plot(data[N+clean+pih+urb+bio+road],     lw=2, label="N c&corr press,incom,humid,urban,bio,road", color='C3')
                    if mov != '': plt.plot(data[N+clean+pih+urb+bio+road+mov], lw=1, label="N c&corr press,incom,humid,urban,bio,road,smoothed", color='black')
                    #if Nfinal != N+clean+pih:
                    #    plt.plot(data[Nfinal], label="N cleaned corrected_pih smoothed_%s" % config['clean']['smooth'], color='C0')
                    ax.fill_between(data.index, data[Nfinal]-data[Nfinal+'_err'], data[Nfinal]+data[Nfinal+'_err'], color='black', alpha=0.2, lw=0)
                    #plot_errorbar(data.index, data[Nfinal], xerr=None, yerr=data[Nfinal+err], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='black', ms=3, mew=0)

                else:
                    plt.plot(data.index[0], 0)

                if 'N_neighbors' in data.columns:
                    #JJ
                    if not data['N_neighbors'].isnull().all():
                        plt.plot(data['N_neighbors'], label='N_neighbors: spatially smoothed', drawstyle='steps')

                plt.legend()

            if Nth:

                with Fig(fig, title='Thermal neutrons', ylabel='Thermal Neutrons (in cph)') as ax:
                    if not data[N].isnull().all():
                        plt.plot(data[N+th+clean],     label="N_thermal cleaned", ls='', marker='o', ms=3.5, markeredgewidth=0, color='whitesmoke')
                        plt.plot(data[N+th+clean+pih], label="N_thermal c&corr press,incom,humid", lw=2, color='black')
                        plt.legend()

                with Fig(fig, title='Thermal and Epithermal neutrons corrected for p,i,h (sea level)', ylabel='Neutrons (in cph)') as ax:
                    if not data[N].isnull().all():
                        plt.plot(data[N+clean+pih], label="Epithermal", lw=2, color='C0')
                        plt.plot(data[N+th+clean+pih],     label="Thermal", lw=2, color='C1')
                        plt.legend()

        #            if N+'_std' in data.columns:
        #                ax.fill_between(data.index, data[N]-data[N+'_std'], data[N]+data[N+'_std'], alpha=0.25, lw=0, step='pre')

#            if Nth:
#                with Fig(fig, title='Thermal neutron time series, corrected', ylabel='Thermal neutrons (in cph)') as ax:
#                    if not data[Nth].isnull().all():
#                        plt.plot(data[Nth], label="N_thermal", drawstyle='steps')


            # Corrections
            with Fig(fig, title='Corrections', ylabel='-') as ax:
                if not data[N].isnull().all():
                    plt.plot(data['correct_p'], label='air pressure', color='grey')
                    plt.plot(data['correct_i'], label='incoming neutron flux', color='C1')
                    plt.plot(data['correct_h'], label='air humidity', color='C0')
                    if urb != '': plt.plot(data['correct_urban'], label='urban area', lw=5, color='C6')
                    if bio != '': plt.plot(data['correct_bio'], label='biomass', color='C2')
                    if road!= '': plt.plot(data['correct_road'], label='road effect', color='black', lw=1)
                    plt.legend()
                else:
                    plt.plot(data.index[0], 0)

                if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None'and lat in data.columns:
                    with Fig(fig, title='Road parameters', ylabel='-') as ax:
                        plt.plot(data['road_width'], label='road_width', marker='o', ls='', ms=2)
                        plt.plot(data['road_moisture'], label='road_moisture', marker='o', ls='', ms=2)
                        plt.legend()

            if data.bd.std().mean() > 0.001:
                with Fig(fig, title='Soil properties', ylabel='-') as ax:
                    plt.plot(data['bd'], label='bulk density', marker='o', ls='', ms=2)
                    plt.plot(data['lw'], label='lattice water', marker='o', ls='', ms=2)
                    plt.plot(data['owe'], label='Organic water equivalent', marker='o', ls='', ms=2)
                    plt.legend()

            # Air properties
            with Fig(fig, layout=(2,2)) as axes:
                if not data[p].isnull().all():
                    with Fig(ylabel='Air pressure (in mbar)'):
                        plt.plot(data[p], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=1: axes[1].axis('off')

                if not data[T].isnull().all():
                    with Fig(ylabel='Air temperature (in $^\circ$C)'):
                        plt.plot(data[T], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=2: axes[2].axis('off')

                if not data[rh].isnull().all():
                    with Fig(ylabel='Relative humidity (in %)'):
                        plt.plot(data[rh], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=3: axes[3].axis('off')

                if not data[ah].isnull().all():
                    with Fig(ylabel='Absolute humidity (in g/m$^3$)'):
                        plt.plot(data[ah], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=4: axes[4].axis('off')

            # Moisture
            if config['conversion']['N2sm_method'] != 'None':
                if not data[sm].isnull().all():


#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)) as ax:
#                        plt.plot(data[smg], color='black', ls='', marker='o', ms=4, markeredgewidth=0)
                        #ax.fill_between(data.index, data[smg]+data[smg+err+'_low'], data[smv]+data[smg+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
#                        plot_errorbar(data.index, data[smg], xerr=None, yerr=[-data[smg+err+'_low'], data[smg+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='black', ms=3, mew=0)
#                        plt.yticks(np.arange(0,1.1, 0.1))

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content', ylim=(0,1)) as ax:
#                        plt.plot(data[smv], color='black', ls='', marker='o', ms=4, markeredgewidth=0)
                        #ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
#                        plot_errorbar(data.index, data[smv], xerr=None, yerr=[-data[smv+err+'_low'], data[smv+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='None', ms=3, mew=1)
#                        plt.yticks(np.arange(0,1.1, 0.1))
                    if config['clean']['smooth']:
                        label_str = '%s resolution, $\pm$%0.f smoothing window' % (tres_str, int(config['clean']['smooth'])/2)
                    else:
                        label_str = '%s resolution' % (tres_str)

                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content (in cm$^3$/cm$^3$)', ylim=(0,1)) as ax:
                        ax.fill_between(data.index, data[smv], 0, color='C0', alpha=0.05, lw=0, step='post')
                        plt.plot(data_orig[smv], ls='', color='C0', alpha=0.3, marker='o', ms=2, markeredgewidth=0, label='%s resolution' % tres_orig_str)
                        plt.plot(data[smv], color='C0', drawstyle='steps-post', lw=2, label=label_str)
                        #ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
                        #plot_errorbar(data.index, data[smv], xerr=None, yerr=[-data[smv+err+'_low'], data[smv+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='None', ms=3, mew=1)
                        plt.yticks(np.arange(0,1.1, 0.1))
                        plt.legend(ncol=2, loc='upper center')
                        #ax.set_xlim(str2dt('2020-03-01'),str2dt('2020-09-01'))

                    with Fig(fig, title='Error of the volumetric water content', ylabel='Error (in cm$^3$/cm$^3$)', #) as ax:
                        ylim=(-1.1*data[smv+err+'_upp'].max(), 1.1*data[smv+err+'_upp'].max())) as ax:
                        plt.plot(data[smv+err+'_low'], drawstyle='steps', color='C3', lw=1)
                        plt.plot(data[smv+err+'_upp'], drawstyle='steps', color='C2', lw=1)
                        ax.fill_between(data.index, data[smv+err+'_low'], 0, color='C3', alpha=0.2, lw=0, step='pre')
                        ax.fill_between(data.index, data[smv+err+'_upp'], 0, color='C2', alpha=0.2, lw=0, step='pre')

#                        data[smv+'_int'] = data[smv].interpolate()
#                        data[smv+err+'_low_int'] = data[smv+err+'_low'].interpolate()
#                        data[smv+err+'_upp_int'] = data[smv+err+'_upp'].interpolate()
#                        data = moving_avg(data, smv+'_int', window=15, err=False, std=False)
#                        data = moving_avg(data, smv+err+'_low_int', window=15, err=False, std=False)
#                        data = moving_avg(data, smv+err+'_upp_int', window=15, err=False, std=False)

#                        plt.plot(data[smv+'_int'+mov], color='black', alpha=0.3, drawstyle='steps')
                        #plt.plot(data[smv+'_int'+mov]+data[smv+err+'_low_int'+mov], color='red')
                        #plt.plot(data[smv+'_int'+mov]+data[smv+err+'_upp_int'+mov], color='red')
#                        ax.fill_between(data.index, data[smv+'_int'+mov]+data[smv+err+'_low_int'+mov],
#                                                    data[smv+'_int'+mov]+data[smv+err+'_upp_int'+mov],
#                                                    color='black', alpha=0.1, lw=0, step='pre')

                        #plt.plot(data[smv] +data[smv+err+'_low'], color='C1', ls='', marker='o', ms=2, markeredgewidth=0)
                        #plt.plot(data[smv] +data[smv+err+'_upp'], color='C1', ls='', marker='o', ms=2, markeredgewidth=0)

#                        plt.errorbar(data.index, data[smv], yerr=[data[smv] +data[smv+err+'_low'], data[smv] +data[smv+err+'_upp']],
#                                     fmt='s', ecolor='black', elinewidth=1, capsize=0, mfc='black', mec='black', ms=4, mew=2)

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)):
#                        plt.plot(data[smg], color='black', drawstyle='steps')
#                        plt.plot(data[smg] +data[smg+err+'_low'], color='C1', drawstyle='steps')
#                        plt.plot(data[smg] +data[smg+err+'_upp'], color='C1', drawstyle='steps')


#                        ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
                        #plt.plot(data[smv] -data[smv+err], color='C2')
                        #plt.plot(data[smv] +data[smv+err], color='C2')
#                        if calibration:
#                            plt.plot([datac.index.min(), datac.index.max()], [float(config['conversion']['measured_sm'])*100, float(config['conversion']['measured_sm'])*100], color='red', lw=10, alpha=0.5)
#                            plt.text(datac.index.min(), float(config['conversion']['measured_sm'])*105, 'calibration', rotation=90, va='bottom', color='red', alpha=0.5)

#                    datax = data.copy()
#                    datax[smv] = datax[smv].interpolate()
                    #datax = moving_avg(datax, smv, window=15, err=False, std=False)
#                    datax[smv+err+'_low'] = datax[smv+err+'_low'].interpolate()
                    #datax = moving_avg(datax, smv+err+'_low', window=15, err=False, std=False)
#                    datax[smv+err+'_upp'] = datax[smv+err+'_upp'].interpolate()
                    #datax = moving_avg(datax, smv+err+'_upp', window=15, err=False, std=False)
#                    datay = minly(datax,5)
#                    datay = moving_avg(datay, smv, window=6, err=False, std=False)
#                    datay = moving_avg(datay, smv+err+'_low', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smv+err+'_upp', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg, window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg+err+'_low', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg+err+'_upp', window=6, err=False, std=False)

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content', ylim=(0,1)) as ax:
#                        plt.plot(datay[smv+mov], color='black', drawstyle='steps')
                        #plt.plot(datay[smv+mov] +datay[smv+err+'_low'+mov], color='C1', drawstyle='steps')
                        #plt.plot(datay[smv+mov] +datay[smv+err+'_upp'+mov], color='C1', drawstyle='steps')
#                        ax.fill_between(datay.index, datay[smv+mov]+datay[smv+err+'_low'+mov], datay[smv+mov]+datay[smv+err+'_upp'+mov], color='C0', alpha=0.4, lw=0, step='pre')

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)) as ax:
#                        plt.plot(datay[smg+mov], color='black', drawstyle='steps')
#                        ax.fill_between(datay.index, datay[smg+mov]+datay[smg+err+'_low'+mov], datay[smg+mov]+datay[smg+err+'_upp'+mov], color='C0', alpha=0.4, lw=0, step='pre')

                    # hist
                    print('Histogram... ', end='')
                    with Fig(fig, title='PDF '+sm, ylabel='probability density', time_series=False, size=(10,5)) as ax:
                        ax.hist(data[sm].dropna().values, bins=100, density=True, histtype='stepfilled')
                        ax.set_xlim(0,1)
                        ax.set_xticks(np.arange(0,1,0.1))
                        ax.set_xlabel('vol. soil moisture (in m$^3$/m$^3$)')
                        #ax.set_ylabel('probability density')
                        #ax.set_title('Soil moisture PDF') ;
                        ax.grid(color='black', alpha=0.1)


            # More plots
            for var in xsplit(config['output']['PDF_plots']):

                ## Footprint
                if var == 'footprint':

                    with Fig(fig, layout=(3,1)) as axes:

                        var = 'footprint_radius'
                        with Fig(title='Footprint', ylabel='Radius (in m)', ylim=(0,350)) as ax:
                            ax.fill_between(data.index, data[var], 0, color='C2', alpha=0.3, lw=0, step='pre')
                            plt.plot(data[var], color='C2', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(color='black', alpha=0.3)

                        var = 'footprint_depth'
                        with Fig(ylabel='Depth (in cm)', ylim=(-80,0)) as ax:
                            ax.fill_between(data.index, -data[var], 0, color='sienna', alpha=0.3, lw=0, step='pre')
                            plt.plot(-data[var], color='sienna', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(color='black', alpha=0.3)

                        var = 'footprint_volume'
                        with Fig(ylabel='Volume (in $10^3$ m$^3$)', ylim=(0,120)) as ax:
                            ax.fill_between(data.index, data[var], 0, color='C0', alpha=0.3, lw=0, step='pre')
                            plt.plot(data[var], color='C0', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(color='black', alpha=0.3)

                ## Plot gridded data (EXPERIMENTAL)
                elif var == 'map_grid':

                    var = sm
                    if lat != '' and lat in data.columns:
                        datax = data[[lon,lat,var]].dropna()
                        if len(datax)==0:
                            raise ValueError('No records with valid sm left after processing, stop printing')

                        ## Deprecated:
                        # import pyproj
                        # xs, ys = pyproj.transform(pyproj.Proj("epsg:4326"), pyproj.Proj("epsg:31468"), datax[lat].values, datax[lon].values)
                        ## New pyproj2 syntax:
                        from pyproj import Transformer
                        xs, ys = Transformer.from_crs("epsg:4326", "epsg:31468").transform(datax[lat].values, datax[lon].values)

                        datax['y'] = xs
                        datax['x'] = ys

                        from scipy.interpolate import griddata
                        # data coordinates and values
                        x = datax.x.values - datax.x.min()
                        y = datax.y.values - datax.y.min()
                        z = datax[var].values

                        # config
                        print('XY plot... ', end='')
                        resolution = 70
                        smrange = (0,0.5)
                        contour_levels = np.arange(0.05,0.525,0.025) #20

                        xrange = x.max()-x.min()
                        yrange = y.max()-y.min()
                        if xrange==0: xrange=1
                        if yrange==0: yrange=1
                        xpad = 0.05*xrange
                        ypad = 0.10*yrange

                        mysize = (10,10/xrange*yrange)
                        if mysize[1] > 20:
                            mysize = (10*xrange/yrange,10)
                        with Fig(fig, title='Map '+var, xlabel='Easting (in m)', ylabel='Northing (in m)', size=mysize, time_series=False) as ax:
                            q = ax.scatter( x, y, c=z, cmap='Spectral', vmin=smrange[0], vmax=smrange[1])
                            ax.set_aspect(1)
                            ax.set_xlim(x.min()-xpad,x.max()+xpad)
                            ax.set_ylim(y.min()-ypad,y.max()+ypad)
                            #ax.set_xlabel('Easting [m]')
                            #ax.set_ylabel('Northing [m]')
                            #ax.set_title('Soil Moisture Map') ;
                            cb = plt.colorbar(q, ax=ax, shrink=0.6, pad=0.03, aspect=30)
                            cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                ## Plot with tiles (EXPERIMENTAL)
                elif var == 'map_tiles':

                    var = sm
                    # TODO Those params should go to the config file
                    todo_config_zoom = 16
                    todo_config_map = 'satellite'
                    todo_config_markersize = 15

                    print('Latlon plot with tiles...', end='')
                    datall = data.dropna(subset=[sm]).copy()

                    if todo_config_map == 'satellite':
                        cimgt.QuadtreeTiles.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.QuadtreeTiles() # spoofed, downloaded street map
                    elif todo_config_map == 'OSM':
                        cimgt.OSM.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.OSM() # spoofed, downloaded street map
                    else:
                        todo_config_map = None

                    lonpad = 0.1*(datall[lon].max()-datall[lon].min())
                    latpad = 0.3*(datall[lat].max()-datall[lat].min())

                    zoom = todo_config_zoom #18-int(15*(data[lon].max()-data[lon].min())) # 0->18, 0.5->16, 1->14
                    # Do not choose >=17 for more than a small field site! Python will crash.
                    #print("(zoom %i) " % zoom, end='')

                    with Fig(fig, title='Map '+var, xlabel='lon', ylabel='lat', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:
                        ax.scatter( datall[lon], datall[lat], c='black', transform=ccrs.PlateCarree(), s=todo_config_markersize*2)
                        q = ax.scatter( datall[lon], datall[lat], c=datall[sm], transform=ccrs.PlateCarree(),
                                        s=todo_config_markersize, cmap='Spectral', vmin=smrange[0], vmax=smrange[1])
                        ax.set_xlim(datall[lon].min()-lonpad, datall[lon].max()+lonpad)
                        ax.set_ylim(datall[lat].min()-latpad, datall[lat].max()+latpad)
                        if not todo_config_map is None:
                            ax.add_image(osm_img, zoom)

                        cb = plt.colorbar(q, ax=ax, shrink=0.7, pad=0.03, aspect=30)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)

                ## Plot grid with tiles (EXPERIMENTAL)
                elif var == 'map_grid_tiles':

                    var = sm
                    # TODO Those params should go to the config file
                    todo_config_zoom = 17
                    todo_config_map = 'satellite'
                    todo_config_resolution = 3
                    todo_config_fineresolu = 2.1   # Selbitz 2.1 # 4.5 #1.5*0.62*4 # Corinna, 1.5*0.62 # Hordorf
                    todo_config_markersize = 100 # 400 #100/4.5 # Corinna, 100 # Hordorf

                    print('Latlon plot with regular grid and tiles...', end='')

                    if todo_config_map == 'satellite':
                        cimgt.QuadtreeTiles.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.QuadtreeTiles() # spoofed, downloaded street map
                    elif todo_config_map == 'OSM':
                        cimgt.OSM.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.OSM() # spoofed, downloaded street map
                    else:
                        todo_config_map = None

                    #print(cprintf('! Internet connection failed. Skipping basemap...','red'))

                    zoom = todo_config_zoom #18-int(15*(data[lon].max()-data[lon].min())) # 0->18, 0.5->16, 1->14
                    # Do not choose >=17 for more than a small field site! Python will crash.
                    #print("(zoom %i) " % zoom, end='')

                    G = data[[lat,lon,smv,rh,T]]
                    G[smv+'_pc'] = G[smv]*100
                    G.loc[:,lat] = fineround(G[lat], todo_config_resolution, todo_config_fineresolu)
                    G.loc[:,lon] = fineround(G[lon], todo_config_resolution, todo_config_fineresolu)
                    G = G.groupby([lat,lon]).mean()
                    G = G.reset_index() # from MultiIndex back to normal columns

                    utmy, utmx = latlon2utm(G[lat], G[lon])
                    G['utmx'] = utmx
                    G['utmy'] = utmy
                    G.to_csv(output_basename+'-grid.csv',
                        date_format = datetime_format,
                        sep = config['output']['CSV_column_sep'].strip('\"'),
                        decimal = config['output']['CSV_decimal'],
                        float_format = '%' + config['output']['CSV_float_format'],
                        na_rep = config['output']['CSV_NaN'])

                    lonpad = 0.1*(data[lon].max()-data[lon].min())
                    latpad = 0.3*(data[lat].max()-data[lat].min())

                    with Fig(fig, title='CRNS aggregated map '+var, size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                        q = ax.scatter( G[lon], G[lat], c=G[smv], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                       cmap='Spectral', vmin=smrange[0], vmax=smrange[1], transform=ccrs.PlateCarree())
                        ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                        ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                        ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                        #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                        ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                        ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                        ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                        if not todo_config_map is None:
                            ax.add_image(osm_img, zoom)

                        cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                        plt.grid(color='black', alpha=0.1)

                    if cc(config, 'output', 'map_grid_tiles_temperature', 'yes'):
                        with Fig(fig, title='Temperature', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                            q = ax.scatter( G[lon], G[lat], c=G[T], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                           cmap='Spectral_r', transform=ccrs.PlateCarree())
                            ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                            ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                            ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                            #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                            ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                            ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                            ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                            if not todo_config_map is None:
                                ax.add_image(osm_img, zoom)

                            cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                            cb.set_label("Temperature (in $^\circ$C)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                    if cc(config, 'output', 'map_grid_tiles_humidity', 'yes'):
                        with Fig(fig, title='Relative air humidity', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                            q = ax.scatter( G[lon], G[lat], c=G[rh], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                           cmap='Spectral', transform=ccrs.PlateCarree())
                            ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                            ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                            ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                            #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                            ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                            ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                            ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                            if not todo_config_map is None:
                                ax.add_image(osm_img, zoom)

                            cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                            cb.set_label("Relative air humidity (in %)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                ## Plot interpolation (EXPERIMENTAL)
                elif var == 'map_interpolation':

                    var = sm
                    # TODO Those params should go to the config file
                    xi = np.arange(x.min()-xpad,x.max()+xpad, resolution)
                    yi = np.arange(y.min()-ypad,y.max()+ypad, resolution)
                    xi,yi = np.meshgrid(xi,yi)
                    # interpolate
                    zi = griddata((x,y),z,(xi,yi), method='linear') #, fill_value=0.3
                    # plot
                    with Fig(fig, title='Lin. Interpolation on %.0f m grid, %s' % (resolution, var), xlabel='Easting (in m)', ylabel='Northing (in m)', size=(10,10/(xrange/yrange)**2), time_series=False) as ax:
                        q = ax.contourf(xi, yi, zi, cmap='Spectral', vmin=smrange[0], vmax=smrange[1], levels=contour_levels, extend='both', corner_mask=False)
                        #q = plt.imshow(zi, cmap='Spectral', vmin=0.1, vmax=0.6, interpolation='gaussian', origin='lower', extent=(x.min()-3*xpad,x.max()+3*xpad, y.min()-3*ypad,y.max()+3*ypad))
                        ax.plot(x, y,'k.', ms=1.5)
                        #ax.set_xlabel('Easting [m]', fontsize=10)
                        #ax.set_ylabel('Northing [m]', fontsize=10)
                        #ax.set_title('Soil moisture linear interpolation')
                        cb = plt.colorbar(q, ax=ax, shrink=0.3, pad=0.03, aspect=30) #, ticks=np.arange(0,0.5,0.05)) #ticks=np.arange(np.ceil(np.nanmin(zi)*20)/20,np.nanmax(zi),0.05)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                        plt.grid(color='black', alpha=0.1)

                elif var in data.columns:
                    with Fig(fig, title=var, ylabel=var):
                        if not data[var].isnull().all():
                            plt.plot(data[var], ls='', marker='o', ms=5, markeredgewidth=0) #drawstyle='steps')
        # Done
        print('OK')
        file_save_msg(figure_name, "Figures", end="\n")

    # Make KML if lat/lon is provided
    if lat in data.columns and lon in data.columns:

        if yesno2bool(config['output']['make_KML_neutrons']):
            if not data[Nfinal].isnull().all():
                print('| Creating_KML for %s... ' % Nfinal)
                collim = xsplit(config['output']['KML_neutrons_range'])
                if len(collim)<=1: collim = None
                save_KML(data, Nfinal, lat, lon, alt='Alt', file=output_basename+'-'+config['correction']['new_neutron_column'], format_str='%i', cmap='Spectral', collim=collim)#(3000,9000)) # long track: 5000,11000
                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'.kml', 'KML', end="\n")

                save_KML(data, 'RecordNum', lat, lon, alt='Alt', file=output_basename+'-ID', format_str='%i')

#            if Nth:
#                print("\n"+'Creating_KML... ', end='')
#                save_KML(data, Nth, lat, lon, file=output_basename+'-'+config['correction']['new_neutron_column']+'_thermal', format_str='', cmap='Spectral') #, collim=(3000,9000)) # long track: 5000,11000
#                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'_thermal.kml', '', end='')

        if yesno2bool(config['output']['make_KML_sm']):
            if not data[sm].isnull().all():
                print('| Creating_KML for %s... ' % sm)
                collim = xsplit(config['output']['KML_sm_range'])
                if len(collim)<=1: collim = None
                save_KML(data, sm+'_%', lat, lon, alt='Alt', file=output_basename+'-'+sm, format_str='%.0f%%', collim=collim)#(5,80))
                file_save_msg(output_basename+'-'+sm+'.kml', 'KML', end="\n")

        if yesno2bool(config['output']['make_KML_other']):
            other = config['output']['KML_other_column']
            if other in data.columns:
                if not data[other].isnull().all():
                    print('| Creating_KML for %s... ' % other)
                    collim = xsplit(config['output']['KML_other_range'])
                    if len(collim)<=1: collim = None
                    if yesno2bool(config['output']['KML_other_reverse']):
                        cmap = 'Spectral_r'
                    else:
                        cmap = 'Spectral'
                    format_str = '%'+config['output']['KML_other_format'] if config['output']['KML_other_format'] else ''
                    save_KML(data, other, lat, lon, alt='Alt', file=output_basename+'-'+other, format_str=format_str, cmap=cmap, collim=collim)#(5,80))
                    file_save_msg(output_basename+'-'+other+'.kml', 'KML', end="\n")

    plt.close('all')
    print("= Done.")
    return(data)

# Direct call (instead of import)

if __name__ == "__main__":
    param2 = ''
    if len(sys.argv) > 1:
        main(sys.argv.pop(1))
        if len(sys.argv) > 1:
            param2 = sys.argv.pop(1)
    else:
        main()

    if param2 != '--nonstop':
        print("Press any key to close.", end='')
        input()
